from __future__ import annotations

import datetime as dt
import enum
import urllib.parse
from collections.abc import AsyncIterator, Collection, Iterable, Sequence
from pathlib import Path
from typing import Any, ClassVar, Protocol, TypedDict, TypeVar

import attrs
from typing_extensions import NotRequired
from yarl import URL

from . import http, manager_ctx, matchers, pkg_archives, pkg_models
from . import results as R
from .catalogue.cataloguer import CatalogueEntry
from .config import GlobalConfig
from .definitions import Defn, SourceMetadata
from .utils import file_uri_to_path, gather, run_in_thread


class FolderHashCandidate(Protocol):  # pragma: no cover
    @property
    def name(self) -> str: ...

    def hash_contents(self, __method: matchers.AddonHashMethod) -> str: ...


TFolderHashCandidate = TypeVar('TFolderHashCandidate', bound=FolderHashCandidate)


class HeadersIntent(enum.IntEnum):
    Download = enum.auto()


class Resolver(Protocol):  # pragma: no cover
    metadata: ClassVar[SourceMetadata]
    'Static source metadata.'

    requires_access_token: ClassVar[str | None]
    'Access token key or ``None``.'

    archive_opener: ClassVar[pkg_archives.ArchiveOpener | None]
    'Alternative archive opener to use supporting e.g. non-standard archive formats or layouts.'

    def __init__(self, manager_ctx: manager_ctx.ManagerCtx) -> None: ...

    @classmethod
    def get_alias_from_url(cls, url: URL) -> str | None:
        "Attempt to extract a ``Defn`` alias from a given URL."
        ...

    async def make_request_headers(
        self, intent: HeadersIntent | None = None
    ) -> dict[str, str] | None:
        "Create headers for resolver HTTP requests."
        ...

    async def resolve(
        self, defns: Sequence[Defn]
    ) -> dict[Defn, pkg_models.Pkg | R.ManagerError | R.InternalError]:
        "Resolve multiple ``Defn``s into packages."
        ...

    async def resolve_one(self, defn: Defn, metadata: Any) -> pkg_models.Pkg:
        "Resolve a ``Defn`` into a package."
        ...

    async def get_changelog(self, uri: URL) -> str:
        "Retrieve a changelog from a URI."
        ...

    async def get_folder_hash_matches(
        self, candidates: Collection[TFolderHashCandidate]
    ) -> Iterable[tuple[Defn, frozenset[TFolderHashCandidate]]]:
        "Find ``Defn``s from folder fingerprint."
        ...

    @classmethod
    def catalogue(cls, web_client: http.ClientSessionType) -> AsyncIterator[CatalogueEntry]:
        "Enumerate add-ons from the source."
        ...


class PkgCandidate(TypedDict):
    "A subset of the ``Pkg`` constructor's kwargs."

    id: str
    slug: str
    name: str
    description: str
    url: str
    download_url: str
    date_published: dt.datetime
    version: str
    changelog_url: str
    deps: NotRequired[list[pkg_models.PkgDep]]


class BaseResolver(Resolver, Protocol):
    _manager_ctx: manager_ctx.ManagerCtx

    archive_opener = None

    def __init__(self, manager_ctx: manager_ctx.ManagerCtx) -> None:
        self._manager_ctx = manager_ctx

    __orig_init = __init__

    def __init_subclass__(cls) -> None:
        # ``Protocol`` clobbers ``__init__`` on Python < 3.11.
        if cls.__init__ is _DummyResolver.__init__:
            cls.__init__ = cls.__orig_init

    @classmethod
    def _get_access_token(cls, global_config: GlobalConfig, name: str | None = None) -> str | None:
        name = name or cls.requires_access_token
        if name is not None:
            return getattr(global_config.access_tokens, name, None)

    @classmethod
    def get_alias_from_url(cls, url: URL) -> str | None:
        return None

    async def make_request_headers(
        self, intent: HeadersIntent | None = None
    ) -> dict[str, str] | None:
        return None

    async def resolve(
        self, defns: Sequence[Defn]
    ) -> dict[Defn, pkg_models.Pkg | R.ManagerError | R.InternalError]:
        results = await gather((self.resolve_one(d, None) for d in defns), R.resultify_async_exc)
        return dict(zip(defns, results))

    async def resolve_one(self, defn: Defn, metadata: Any) -> pkg_models.Pkg:
        extraneous_strategies = defn.strategies.filled_strategies.keys() - self.metadata.strategies
        if extraneous_strategies:
            raise R.PkgStrategiesUnsupported(extraneous_strategies)

        pkg_candidate = await self._resolve_one(defn, metadata)
        return pkg_models.Pkg(
            **pkg_candidate,
            source=self.metadata.id,
            options=pkg_models.PkgOptions(
                **attrs.asdict(defn.strategies, value_serializer=lambda t, a, v: bool(v))
            ),
        )

    async def _resolve_one(self, defn: Defn, metadata: Any) -> PkgCandidate:
        "Resolve a ``Defn`` into a package."
        ...

    async def get_changelog(self, uri: URL) -> str:
        match uri.scheme:
            case 'data' if uri.raw_path.startswith(','):
                return urllib.parse.unquote(uri.raw_path[1:])

            case 'file':
                return await run_in_thread(Path(file_uri_to_path(str(uri))).read_text)(
                    encoding='utf-8'
                )

            case 'http' | 'https':
                async with self._manager_ctx.web_client.get(
                    uri,
                    expire_after=http.CACHE_INDEFINITELY,
                    headers=await self.make_request_headers(),
                    raise_for_status=True,
                ) as response:
                    return await response.text()

            case _:
                raise ValueError('Unsupported URI with scheme', uri.scheme)

    async def get_folder_hash_matches(
        self, candidates: Collection[TFolderHashCandidate]
    ) -> Iterable[tuple[Defn, frozenset[TFolderHashCandidate]]]:
        return []

    @classmethod
    async def catalogue(cls, web_client: http.ClientSessionType) -> AsyncIterator[CatalogueEntry]:
        return
        yield


class _DummyResolver(Resolver):
    pass



from __future__ import annotations

from collections.abc import Callable
from pathlib import Path
from typing import Any

import aiohttp
import aiohttp.web
import pytest
from aresponses import ResponsesMockServer
from aresponses.errors import NoRouteFoundError
from loguru import logger

from instawow.config import GlobalConfig, ProfileConfig
from instawow.http import init_web_client
from instawow.manager_ctx import ManagerCtx, contextualise
from instawow.pkg_management import PkgManager
from instawow.wow_installations import _DELECTABLE_DIR_NAMES, Flavour

from .fixtures.http import ROUTES


def pytest_addoption(parser: pytest.Parser):
    parser.addoption('--iw-no-mock-http', action='store_true')


def should_mock(fn: Callable[..., object]):
    import inspect

    def wrapper(request: pytest.FixtureRequest):
        if request.config.getoption('--iw-no-mock-http'):
            return None
        elif any(m.name == 'iw_no_mock_http' for m in request.node.iter_markers()):
            return None

        args = (request.getfixturevalue(p) for p in inspect.signature(fn).parameters)
        return fn(*args)

    return wrapper


@pytest.fixture
def caplog(caplog: pytest.LogCaptureFixture):
    handler_id = logger.add(
        caplog.handler,
        format='{message}',
        level=0,
        filter=lambda record: record['level'].no >= caplog.handler.level,
        enqueue=False,  # Set to 'True' if your test is spawning child processes.
    )
    yield caplog
    logger.remove(handler_id)


class _StrictResponsesMockServer(ResponsesMockServer):
    async def _find_response(self, request: aiohttp.web.Request):
        response = await super()._find_response(request)
        if response == (None, None):
            raise NoRouteFoundError(f'No match found for <{request.method} {request.url}>')
        return response


@pytest.fixture
async def iw_aresponses():
    async with _StrictResponsesMockServer() as server:
        yield server


@pytest.fixture(params=['foo'])
def iw_global_config_values(request: pytest.FixtureRequest, tmp_path: Path):
    return {
        'config_dir': tmp_path / '__config__' / 'config',
        'temp_dir': tmp_path / '__config__' / 'temp',
        'state_dir': tmp_path / '__config__' / 'state',
        'access_tokens': {
            'cfcore': request.param,
            'github': None,
            'wago': None,
            'wago_addons': request.param,
        },
    }


@pytest.fixture(params=[Flavour.Retail])
def iw_config_values(request: pytest.FixtureRequest, tmp_path: Path):
    installation_dir = (
        tmp_path
        / 'wow'
        / next(k for k, v in _DELECTABLE_DIR_NAMES.items() if v['flavour'] is request.param)
    )
    addon_dir = installation_dir / 'interface' / 'addons'
    addon_dir.mkdir(parents=True)
    return {
        'profile': '__default__',
        'addon_dir': addon_dir,
        'game_flavour': request.param,
        '_installation_dir': installation_dir,
    }


@pytest.fixture
def iw_config(iw_config_values: dict[str, Any], iw_global_config_values: dict[str, Any]):
    global_config = GlobalConfig.from_values(iw_global_config_values).write()
    return ProfileConfig.from_values({'global_config': global_config, **iw_config_values}).write()


@pytest.fixture(autouse=True)
def _iw_global_config_defaults(
    monkeypatch: pytest.MonkeyPatch,
    iw_global_config_values: dict[str, Any],
):
    monkeypatch.setenv('INSTAWOW_CONFIG_DIR', str(iw_global_config_values['config_dir']))
    monkeypatch.setenv('INSTAWOW_TEMP_DIR', str(iw_global_config_values['temp_dir']))
    monkeypatch.setenv('INSTAWOW_STATE_DIR', str(iw_global_config_values['state_dir']))


@pytest.fixture
async def iw_web_client(iw_config: ProfileConfig):
    async with init_web_client(iw_config.global_config.cache_dir) as web_client:
        yield web_client


@pytest.fixture
def iw_manager_ctx(iw_config: ProfileConfig, iw_web_client: aiohttp.ClientSession):
    contextualise(web_client=iw_web_client)
    return ManagerCtx(iw_config)


@pytest.fixture
def iw_manager(iw_manager_ctx: ManagerCtx):
    return PkgManager(iw_manager_ctx)


@pytest.fixture(autouse=True, params=['all'])
@should_mock
def _iw_mock_aiohttp_requests(
    request: pytest.FixtureRequest, iw_aresponses: _StrictResponsesMockServer
):
    if request.param == 'all':
        routes = ROUTES.values()
    else:
        if not request.param.issubset(ROUTES.keys()):
            raise ValueError('Supplied routes must be subset of all routes')

        routes = (ROUTES[k] for k in ROUTES.keys() & request.param)

    for route in routes:
        iw_aresponses.add(**route.to_aresponses_add_args())



from __future__ import annotations

import os
import subprocess
import sys
from collections.abc import Iterator
from enum import Enum
from functools import cache
from pathlib import Path
from typing import Protocol, TypedDict

from typing_extensions import Self, TypeVar

from .utils import StrEnum, fill

_TEnum = TypeVar('_TEnum', bound=Enum, infer_variance=True)


class _FlavourKeyedEnumMeta(type(Protocol)):  # pragma: no cover
    def __getitem__(self: type[_FlavourKeyedEnum[_TEnum]], __key: str) -> _TEnum: ...


class _FlavourKeyedEnum(Protocol[_TEnum], metaclass=_FlavourKeyedEnumMeta):  # pragma: no cover
    Retail: _TEnum
    VanillaClassic: _TEnum
    Classic: _TEnum
    CataclysmClassic: _TEnum


class Flavour(StrEnum):
    # The current Classic version is always aliased to "classic".
    # The assumption here is that should Classic not be discontinued
    # it will continue to be updated in place so that new Classic versions
    # will inherit the "_classic_" folder.  This means we won't have to
    # migrate Classic profiles either automatically or by requiring user
    # intervention for new Classic releases.
    Retail = 'retail'
    VanillaClassic = 'vanilla_classic'
    Classic = 'classic'
    CataclysmClassic = 'cataclysm_classic'

    @classmethod
    def _missing_(cls, value: object) -> Flavour | None:
        match value:
            # case 'cataclysm_classic':
            #     return cls.Classic
            case _:
                return None

    @classmethod
    def from_flavour_keyed_enum(cls, flavour_keyed_enum: Enum) -> Self:
        return cls[flavour_keyed_enum.name]

    def to_flavour_keyed_enum(self, flavour_keyed_enum: type[_FlavourKeyedEnum[_TEnum]]) -> _TEnum:
        return flavour_keyed_enum[self.name]


class FlavourVersionRange(Enum):
    Retail = (
        range(1_00_00, 1_13_00),
        range(2_00_00, 2_05_00),
        range(3_00_00, 3_04_00),
        range(4_00_00, 4_04_00),
        range(5_00_00, 12_00_00),
    )
    VanillaClassic = (range(1_13_00, 2_00_00),)
    Classic = (range(3_04_00, 4_00_00),)
    CataclysmClassic = (range(4_04_00, 5_00_00),)

    @classmethod
    def _parse_version_string(cls, version_string: str) -> int:
        major, minor, patch = fill(map(int, version_string.split('.')), 0, 3)
        return major * 1_00_00 + minor * 1_00 + patch

    @classmethod
    def from_version(cls, version: int | str) -> Self | None:
        version_number = (
            version if isinstance(version, int) else cls._parse_version_string(version)
        )
        return next((f for f in cls if f.contains(version_number)), None)

    def contains(self, version: int | str) -> bool:
        version_number = (
            version if isinstance(version, int) else self._parse_version_string(version)
        )
        return any(version_number in r for r in self.value)


class _Product(TypedDict):
    code: str
    flavour: Flavour


_DELECTABLE_DIR_NAMES: dict[str, _Product] = {
    '_retail_': {
        'code': 'wow',
        'flavour': Flavour.Retail,
    },
    '_ptr_': {
        'code': 'wowt',
        'flavour': Flavour.Retail,
    },
    '_xptr_': {
        'code': 'wowxptr',
        'flavour': Flavour.Retail,
    },
    '_beta_': {
        'code': 'wow_beta',
        'flavour': Flavour.Retail,
    },
    '_classic_era_': {
        'code': 'wow_classic_era',
        'flavour': Flavour.VanillaClassic,
    },
    '_classic_era_ptr_': {
        'code': 'wow_classic_era_ptr',
        'flavour': Flavour.VanillaClassic,
    },
    # '_classic_era_beta_': {
    #     'code': 'wow_classic_era_beta',
    #     'flavour': Flavour.VanillaClassic,
    # },
    '_classic_': {
        'code': 'wow_classic',
        'flavour': Flavour.Classic,
    },
    '_classic_ptr_': {
        'code': 'wow_classic_ptr',
        'flavour': Flavour.Classic,
    },
    '_classic_beta_': {
        'code': 'wow_classic_beta',
        'flavour': Flavour.CataclysmClassic,
    },
}

ADDON_DIR_PARTS = ('Interface', 'AddOns')
_NORMALISED_ADDON_DIR_PARTS = tuple(map(str.casefold, ADDON_DIR_PARTS))


def _find_mac_installations():
    try:
        possible_installations = subprocess.check_output(
            [
                'mdfind',
                "kMDItemCFBundleIdentifier == 'com.blizzard.worldofwarcraft'",
            ],
            text=True,
            timeout=3,
        ).splitlines()
    except subprocess.TimeoutExpired:
        pass
    else:
        for match in possible_installations:
            installation_path = Path(match).parent
            yield (installation_path, _DELECTABLE_DIR_NAMES.get(installation_path.name))


def find_installations() -> Iterator[tuple[Path, _Product | None]]:
    if sys.platform == 'darwin':
        yield from _find_mac_installations()


def _read_info(info_path: Path):
    with open(info_path, encoding='utf-8') as info:
        header = next(info)
        fields = tuple(f[: f.find('!')] for f in header.rstrip().split('|'))
        return [dict(zip(fields, e.rstrip().split('|'))) for e in info if not e.startswith('#')]


@cache
def _get_installed_versions(outer_installation_path: Path):
    return {
        e['Product']: e['Version'] for e in _read_info(outer_installation_path / '.build.info')
    }


def get_installation_dir_from_addon_dir(path_like: os.PathLike[str] | str) -> Path | None:
    path = Path(path_like)
    tail = tuple(map(str.casefold, path.parts[-3:]))
    if len(tail) == 3 and tail[1:] == _NORMALISED_ADDON_DIR_PARTS:
        return path.parents[1]


def infer_flavour_from_addon_dir(path: os.PathLike[str] | str) -> Flavour | None:
    maybe_installation_dir = get_installation_dir_from_addon_dir(path)
    if maybe_installation_dir:
        try:
            return _DELECTABLE_DIR_NAMES[maybe_installation_dir.name]['flavour']
        except KeyError:
            return None


def get_installation_version_from_addon_dir(path: os.PathLike[str] | str) -> str | None:
    maybe_installation_dir = get_installation_dir_from_addon_dir(path)
    if maybe_installation_dir:
        product = _DELECTABLE_DIR_NAMES.get(maybe_installation_dir.name)
        if product:
            all_versions = _get_installed_versions(maybe_installation_dir.parent)
            return all_versions[product['code']]



from __future__ import annotations

import contextvars as cv
from collections.abc import (
    Collection,
    Mapping,
    Sequence,
)
from contextlib import AbstractAsyncContextManager
from functools import cached_property
from itertools import chain
from typing import TypeAlias

from . import http, pkg_db
from ._sources.cfcore import CfCoreResolver
from ._sources.github import GithubResolver
from ._sources.instawow import InstawowResolver
from ._sources.tukui import TukuiResolver
from ._sources.wago import WagoResolver
from ._sources.wowi import WowiResolver
from .config import ProfileConfig
from .pkg_archives import ArchiveOpener, open_zip_archive
from .plugins import get_plugin_resolvers
from .resolvers import Resolver
from .utils import WeakValueDefaultDictionary


class _DummyLock:
    async def __aenter__(self):
        pass

    async def __aexit__(self, *args: object):
        pass


class _Resolvers(dict[str, Resolver]):
    @cached_property
    def archive_opener_dict(self) -> _ResolverArchiveOpenerDict:
        return _ResolverArchiveOpenerDict(self)

    @cached_property
    def priority_dict(self) -> _ResolverPriorityDict:
        return _ResolverPriorityDict(self)

    @cached_property
    def addon_toc_key_and_id_pairs(self) -> Collection[tuple[str, str]]:
        return [
            (r.metadata.addon_toc_key, r.metadata.id)
            for r in self.values()
            if r.metadata.addon_toc_key
        ]


class _ResolverArchiveOpenerDict(dict[str, ArchiveOpener]):
    def __init__(self, resolvers: _Resolvers) -> None:
        super().__init__(
            (r.metadata.id, r.archive_opener) for r in resolvers.values() if r.archive_opener
        )

    def __missing__(self, key: str) -> ArchiveOpener:
        return open_zip_archive


class _ResolverPriorityDict(dict[str, float]):
    def __init__(self, resolvers: _Resolvers) -> None:
        super().__init__((n, i) for i, n in enumerate(resolvers))

    def __missing__(self, key: str) -> float:
        return float('inf')


_Locks: TypeAlias = Mapping[object, AbstractAsyncContextManager[None]]

_locks = cv.ContextVar[_Locks]('_locks', default=WeakValueDefaultDictionary(_DummyLock))
_web_client = cv.ContextVar[http.ClientSessionType]('_web_client')


def contextualise(
    *,
    locks: _Locks | None = None,
    web_client: http.ClientSessionType | None = None,
) -> None:
    "Set variables for the current context."
    if locks is not None:
        _locks.set(locks)
    if web_client is not None:
        _web_client.set(web_client)


class ManagerCtx:
    "The one true context."

    RESOLVERS: Sequence[type[Resolver]] = [
        GithubResolver,
        CfCoreResolver,
        WowiResolver,
        TukuiResolver,
        InstawowResolver,
        WagoResolver,
    ]
    'Default resolvers.'

    def __init__(
        self,
        config: ProfileConfig,
    ) -> None:
        self.config: ProfileConfig = config

        builtin_resolver_classes = list(self.RESOLVERS)

        for resolver, access_token in (
            (r, getattr(self.config.global_config.access_tokens, r.requires_access_token, None))
            for r in self.RESOLVERS
            if r.requires_access_token is not None
        ):
            if access_token is None:
                builtin_resolver_classes.remove(resolver)

        resolver_classes = chain(
            (r for g in get_plugin_resolvers() for r in g), builtin_resolver_classes
        )
        self.resolvers: _Resolvers = _Resolvers((r.metadata.id, r(self)) for r in resolver_classes)

    @cached_property
    def database(self) -> pkg_db.sa.Engine:
        return pkg_db.prepare_database(self.config.db_uri)

    @property
    def locks(self) -> _Locks:
        "Lock factory used to synchronise async operations."
        return _locks.get()

    @property
    def web_client(self) -> http.ClientSessionType:
        return _web_client.get()



from __future__ import annotations

import contextlib
import sqlite3
from datetime import datetime, timezone

import sqlalchemy
import sqlalchemy.event
import sqlalchemy.exc

sa = sqlalchemy


_VERSION = 0


class _TZDateTime(sa.TypeDecorator[datetime]):
    impl = sa.DateTime
    cache_ok = True

    def process_bind_param(self, value: datetime | None, dialect: object) -> datetime | None:
        if value is not None:
            if not value.tzinfo:
                raise TypeError('tzinfo is required')
            value = value.astimezone(timezone.utc).replace(tzinfo=None)
        return value

    def process_result_value(self, value: datetime | None, dialect: object) -> datetime | None:
        if value is not None:
            value = value.replace(tzinfo=timezone.utc)
        return value

    @property
    def python_type(self) -> type[datetime]:
        return datetime


_metadata = sa.MetaData()


pkg = sa.Table(
    'pkg',
    _metadata,
    sa.Column('source', sa.String, primary_key=True),
    sa.Column('id', sa.String, primary_key=True),
    sa.Column('slug', sa.String, nullable=False),
    sa.Column('name', sa.String, nullable=False),
    sa.Column('description', sa.String, nullable=False),
    sa.Column('url', sa.String, nullable=False),
    sa.Column('download_url', sa.String, nullable=False),
    sa.Column('date_published', _TZDateTime, nullable=False),
    sa.Column('version', sa.String, nullable=False),
    sa.Column('changelog_url', sa.String, nullable=False),
)

pkg_options = sa.Table(
    'pkg_options',
    _metadata,
    sa.Column('any_flavour', sa.Boolean, nullable=False),
    sa.Column('any_release_type', sa.Boolean, nullable=False),
    sa.Column('version_eq', sa.Boolean, nullable=False),
    sa.Column('pkg_source', sa.String, primary_key=True),
    sa.Column('pkg_id', sa.String, primary_key=True),
    sa.ForeignKeyConstraint(
        ['pkg_source', 'pkg_id'],
        ['pkg.source', 'pkg.id'],
        name='fk_pkg_options_pkg_source_and_id',
        ondelete='CASCADE',
    ),
)

pkg_folder = sa.Table(
    'pkg_folder',
    _metadata,
    sa.Column('name', sa.String, primary_key=True),
    sa.Column('pkg_source', sa.String, nullable=False),
    sa.Column('pkg_id', sa.String, nullable=False),
    sa.ForeignKeyConstraint(
        ['pkg_source', 'pkg_id'],
        ['pkg.source', 'pkg.id'],
        name='fk_pkg_folder_pkg_source_and_id',
        ondelete='CASCADE',
    ),
)

pkg_dep = sa.Table(
    'pkg_dep',
    _metadata,
    sa.Column('id', sa.String, primary_key=True),
    sa.Column('pkg_source', sa.String, primary_key=True),
    sa.Column('pkg_id', sa.String, primary_key=True),
    sa.ForeignKeyConstraint(
        ['pkg_source', 'pkg_id'],
        ['pkg.source', 'pkg.id'],
        name='fk_pkg_dep_pkg_source_and_id',
        ondelete='CASCADE',
    ),
)

pkg_version_log = sa.Table(
    'pkg_version_log',
    _metadata,
    sa.Column('version', sa.String, primary_key=True),
    sa.Column('install_time', _TZDateTime, nullable=False, server_default=sa.func.now()),
    sa.Column('pkg_source', sa.String, primary_key=True),
    sa.Column('pkg_id', sa.String, primary_key=True),
)


def _get_version(engine: sa.Engine) -> int:
    with engine.connect() as connection:
        return connection.execute(sa.text('PRAGMA user_version')).scalar_one()


def _migrate(engine: sa.Engine, current_version: int, new_version: int) -> None:
    from ._migrations import MIGRATIONS

    # SQLite migration reference:
    # https://www.sqlite.org/lang_altertable.html#otheralter

    with (
        engine.begin() as transaction,  # Steps 2 and 11.
        contextlib.ExitStack() as exit_stack,
    ):
        # Steps 1 and 12.
        transaction.exec_driver_sql('PRAGMA foreign_keys = OFF')
        exit_stack.callback(lambda: transaction.exec_driver_sql('PRAGMA foreign_keys = ON'))

        for intermediate_version in range(current_version + 1, new_version + 1):
            MIGRATIONS[intermediate_version].upgrade(transaction)

        # Step 10.
        transaction.exec_driver_sql('PRAGMA foreign_key_check')

        # Stamp database with new version.
        transaction.exec_driver_sql(f'PRAGMA user_version = {new_version}')


def _set_fk_pragma(dbapi_connection: sqlite3.Connection, connection_record: object):
    with contextlib.closing(dbapi_connection.cursor()) as cursor:
        cursor.execute('PRAGMA foreign_keys = ON')


def prepare_database(uri: str) -> sa.Engine:
    "Connect to and optionally create or migrate the database."
    engine = sa.create_engine(
        uri,
        # echo=True,
    )
    sa.event.listen(engine, 'connect', _set_fk_pragma)

    current_version = _get_version(engine)
    if current_version != _VERSION:
        _migrate(engine, current_version, _VERSION)
    else:
        _metadata.create_all(engine)

    return engine



from __future__ import annotations

import datetime as dt
from functools import lru_cache

import cattrs
from attrs import field, frozen

from .definitions import Defn, StrategyValues


@lru_cache(1)
def make_db_pkg_converter():
    converter = cattrs.Converter()
    converter.register_structure_hook(dt.datetime, lambda d, _: d)
    return converter


@frozen(kw_only=True)
class PkgOptions:
    any_flavour: bool
    any_release_type: bool
    version_eq: bool


@frozen(kw_only=True)
class PkgFolder:
    name: str


@frozen(kw_only=True)
class PkgDep:
    id: str


@frozen(kw_only=True)
class PkgLoggedVersion:
    version: str
    install_time: dt.datetime


@frozen(kw_only=True, eq=False)
class Pkg:
    source: str
    id: str
    slug: str
    name: str
    description: str
    url: str
    download_url: str
    date_published: dt.datetime
    version: str
    changelog_url: str
    options: PkgOptions  # pkg_options
    folders: list[PkgFolder] = field(factory=list)  # pkg_folder
    deps: list[PkgDep] = field(factory=list)  # pkg_dep
    logged_versions: list[PkgLoggedVersion] = field(factory=list)  # pkg_version_log

    def to_defn(self) -> Defn:
        return Defn(
            source=self.source,
            alias=self.slug,
            id=self.id,
            strategies=StrategyValues(
                any_flavour=self.options.any_flavour or None,
                any_release_type=self.options.any_release_type or None,
                version_eq=self.version if self.options.version_eq else None,
            ),
        )



from __future__ import annotations

from collections.abc import Mapping
from typing import Protocol

import sqlalchemy as sa


class Migration(Protocol):
    def upgrade(self, connection: sa.Connection) -> None: ...

    def downgrade(self, connection: sa.Connection) -> None: ...


class BaseMigration(Migration):
    def _with_table_op(self, connection: sa.Connection) -> None: ...


MIGRATIONS: Mapping[int, Migration] = {}



from __future__ import annotations

from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from pathlib import Path
from typing import TYPE_CHECKING, Any, Generic, Literal, TypeAlias

from typing_extensions import Never, TypedDict, TypeVar

if TYPE_CHECKING:
    import aiohttp

    ClientSessionType: TypeAlias = aiohttp.ClientSession
else:
    ClientSessionType = type

_USER_AGENT = 'instawow (+https://github.com/layday/instawow)'

_DEFAULT_EXPIRE = 0  # Do not cache by default.

CACHE_INDEFINITELY = -1


_TProgress = TypeVar('_TProgress', bound=str)
_TProgressCtx = TypeVar('_TProgressCtx', bound='ProgressCtx[Any]', default=Never)


class ProgressCtx(TypedDict, Generic[_TProgress]):
    report_progress: _TProgress


class GenericDownloadTraceRequestCtx(ProgressCtx[Literal['generic']]):
    label: str


TraceRequestCtx: TypeAlias = GenericDownloadTraceRequestCtx | _TProgressCtx | None


@asynccontextmanager
async def init_web_client(
    cache_dir: Path | None, *, no_cache: bool = False, **kwargs: Any
) -> AsyncIterator[ClientSessionType]:
    import ssl

    import truststore
    from aiohttp import ClientSession, ClientTimeout, TCPConnector

    connector = TCPConnector(limit_per_host=10, ssl=truststore.SSLContext(ssl.PROTOCOL_TLS_CLIENT))

    kwargs = {
        'connector': connector,
        'headers': {'User-Agent': _USER_AGENT},
        'trust_env': True,  # Respect the ``http(s)_proxy`` env var
        'timeout': ClientTimeout(connect=60, sock_connect=10, sock_read=20),
        **kwargs,
    }

    if cache_dir is not None:
        from aiohttp_client_cache import CacheBackend
        from aiohttp_client_cache.session import CachedSession

        from ._cache import make_cache

        cache_backend = CacheBackend(
            allowed_codes=(200, 206),
            allowed_methods=('GET', 'POST'),
            expire_after=_DEFAULT_EXPIRE,
            include_headers=True,
        )
        with make_cache(cache_dir) as cache:
            cache_backend.responses = cache_backend.redirects = cache
            if no_cache:
                cache_backend.disabled = True

            async with CachedSession(cache=cache_backend, **kwargs) as client_session:
                yield client_session

    else:
        async with ClientSession(**kwargs) as client_session:
            yield client_session



from __future__ import annotations

import asyncio
import json
import sys
from contextlib import suppress
from functools import partial
from typing import cast

import anyio.from_thread
import anyio.to_thread
import toga
import toga.constants
import toga.style.pack
from loguru import logger

from instawow.utils import StrEnum

from . import json_rpc_server


class _TogaSimulateKeypressAction(StrEnum):
    ToggleSearchFilter = 'toggleSearchFilter'
    ActivateViewInstalled = 'activateViewInstalled'
    ActivateViewReconcile = 'activateViewReconcile'
    ActivateViewSearch = 'activateViewSearch'


class InstawowApp(toga.App):
    def __init__(self, debug: bool, **kwargs) -> None:
        self.__debug = debug

        super().__init__(
            formal_name='instawow-gui',
            app_id='org.instawow.instawow_gui',
            app_name='instawow_gui',
            icon='resources/instawow_gui',
            **kwargs,
        )

    def startup(self, **kwargs) -> None:
        self.main_window = toga.MainWindow(title=self.formal_name, size=(800, 600))

        if sys.platform == 'win32':
            import ctypes

            # Enable high DPI support.
            ctypes.windll.user32.SetProcessDPIAware()

        self.main_window.content = web_view = toga.WebView(style=toga.style.pack.Pack(flex=1))

        if self.__debug:
            with suppress(AttributeError):
                web_view._impl.native.configuration.preferences.setValue(
                    True, forKey='developerExtrasEnabled'
                )

        if sys.platform == 'win32':
            web_view_impl = web_view._impl

            def configure_webview2(sender, event_args):
                if event_args.IsSuccess:
                    web_view_impl.native.CoreWebView2.Settings.AreBrowserAcceleratorKeysEnabled = (
                        False
                    )

            web_view_impl.native.CoreWebView2InitializationCompleted += configure_webview2

        def dispatch_js_keyboard_event(command: toga.Command, **kwargs):
            event_args = json.dumps(
                {'detail': {'action': kwargs['action']}, 'bubbles': True, 'cancelable': True}
            )
            web_view.evaluate_javascript(
                f'document.dispatchEvent(new CustomEvent("togaSimulateKeypress", {event_args}));'
            )

            return True

        self.commands.add(
            toga.Command(
                partial(
                    dispatch_js_keyboard_event,
                    action=_TogaSimulateKeypressAction.ToggleSearchFilter,
                ),
                text='Toggle Search Filter',
                shortcut=toga.Key.MOD_1 + toga.Key.G,
                group=cast(toga.Group, toga.Group.EDIT),
                section=20,
                order=10,
            ),
        )
        self.commands.add(
            toga.Command(
                partial(
                    dispatch_js_keyboard_event,
                    action=_TogaSimulateKeypressAction.ActivateViewInstalled,
                ),
                text='Installed',
                shortcut=toga.Key.MOD_1 + toga.Key.L,
                group=cast(toga.Group, toga.Group.WINDOW),
                section=20,
                order=10,
            ),
        )
        self.commands.add(
            toga.Command(
                partial(
                    dispatch_js_keyboard_event,
                    action=_TogaSimulateKeypressAction.ActivateViewReconcile,
                ),
                text='Unreconciled',
                group=cast(toga.Group, toga.Group.WINDOW),
                section=20,
                order=20,
            ),
        )
        self.commands.add(
            toga.Command(
                partial(
                    dispatch_js_keyboard_event,
                    action=_TogaSimulateKeypressAction.ActivateViewSearch,
                ),
                text='Search',
                shortcut=toga.Key.MOD_1 + toga.Key.F,
                group=cast(toga.Group, toga.Group.WINDOW),
                section=20,
                order=30,
            ),
        )

        async def startup():
            async with anyio.from_thread.BlockingPortal() as portal:

                def run_json_rpc_server():
                    async def run():
                        web_app = await json_rpc_server.create_app((self.main_window, portal))
                        server_url, serve_forever = await json_rpc_server.run_app(web_app)

                        logger.debug(f'JSON-RPC server running on {server_url}')

                        set_server_url = partial(setattr, web_view, 'url')
                        portal.call(set_server_url, str(server_url))

                        await serve_forever()

                    # We don't want to inherit the parent thread's event loop policy,
                    # i.e. the rubicon loop on macOS.
                    policy = asyncio.DefaultEventLoopPolicy()
                    policy.new_event_loop().run_until_complete(run())

                await anyio.to_thread.run_sync(run_json_rpc_server)

        main_task = self.loop.create_task(startup())

        def on_exit(app: toga.App, **kwargs):
            return main_task.cancel()

        self.on_exit = on_exit

        self.main_window.show()



from __future__ import annotations

from collections.abc import Iterable
from functools import partial
from typing import Literal

import attrs
from typing_extensions import Self
from yarl import URL

from .utils import StrEnum, fauxfrozen


class Strategy(StrEnum):
    AnyFlavour = 'any_flavour'
    AnyReleaseType = 'any_release_type'
    VersionEq = 'version_eq'


class ChangelogFormat(StrEnum):
    Html = 'html'
    Markdown = 'markdown'
    Raw = 'raw'


@fauxfrozen
class SourceMetadata:
    id: str
    name: str
    strategies: frozenset[Strategy]
    changelog_format: ChangelogFormat
    addon_toc_key: str | None


@fauxfrozen
class StrategyValues:
    any_flavour: Literal[True, None] = None
    any_release_type: Literal[True, None] = None
    version_eq: str | None = None

    initialised = True

    @property
    def filled_strategies(self) -> dict[Strategy, object]:
        return {Strategy(p): v for p, v in attrs.asdict(self).items() if v is not None}


@fauxfrozen
class _UninitialisedStrategyValues(StrategyValues):
    initialised = False


_STRATEGY_SEP = ','
_STRATEGY_VALUE_SEP = '='


@fauxfrozen
class Defn:
    source: str
    alias: str
    id: str | None = None
    strategies: StrategyValues = _UninitialisedStrategyValues()

    @classmethod
    def from_uri(
        cls,
        uri: str,
        *,
        known_sources: Iterable[str],
        allow_unsourced: bool,
    ) -> Self:
        """Construct a ``Defn`` from a URI."""
        url = URL(uri)

        if url.scheme not in known_sources:
            if not allow_unsourced:
                raise ValueError(f'Unable to extract source from {uri}')

            make_cls = partial(cls, source='', alias=uri)
        else:
            make_cls = partial(cls, source=url.scheme, alias=url.path)

        if url.fragment == _STRATEGY_VALUE_SEP:
            make_cls = partial(make_cls, strategies=StrategyValues())
        elif url.fragment:
            strategy_values = {
                s: v
                for f in url.fragment.split(_STRATEGY_SEP)
                for s, _, v in (f.partition(_STRATEGY_VALUE_SEP),)
                if s
            }

            unknown_strategies = strategy_values.keys() - set(Strategy)
            if unknown_strategies:
                raise ValueError(f'Unknown strategies: {", ".join(unknown_strategies)}')

            make_cls = partial(
                make_cls,
                strategies=StrategyValues(
                    any_flavour=Strategy.AnyFlavour in strategy_values or None,
                    any_release_type=Strategy.AnyReleaseType in strategy_values or None,
                    version_eq=strategy_values.get(Strategy.VersionEq),
                ),
            )

        return make_cls()

    def as_uri(self, include_strategies: bool = False) -> str:
        uri = f'{self.source}:{self.alias}'

        if include_strategies:
            filled_strategies = self.strategies.filled_strategies
            if filled_strategies:
                uri += f"""#{_STRATEGY_SEP.join(
                    s if v is True else f"{s}={v}" for s, v in filled_strategies.items()
                )}"""

        return uri

    def with_default_strategy_set(self) -> Self:
        return attrs.evolve(
            self,
            strategies=StrategyValues(),
        )

    def with_version(self, version: str) -> Self:
        return attrs.evolve(
            self,
            strategies=StrategyValues(**(attrs.asdict(self.strategies) | {'version_eq': version})),
        )
